/***************************************************************************
* Copyright (c) 2000-2004, Future Systems, Inc. / Seoul, Republic of Korea *
* All Rights Reserved.                                                     *
*                                                                          *
* This document contains proprietary and confidential information.  No     *
* parts of this document or the computer program it embodies may be in     *
* any way copied, duplicated, reproduced, translated into a different      *
* programming language, or distributed to any person, company, or          *
* corporation without the prior written consent of Future Systems, Inc.    *
*                              Hyo Sun Hwang                               *
*                372-2 YangJae B/D 6th Floor, Seoul, Korea                 *
*                           +82-2-578-0581 (552)                           *
***************************************************************************/

/*--------------------- [ Version/Command in detais] ---------------------*\
Description : filename.? (CrypLib version 1.00)
			(head file)	(C-source file)

C0000 : Created by Hyo Sun Hwang (hyosun@future.co.kr) 2000/12/31

C0001 : Modified by Hyo Sun Hwang (hyosun@future.co.kr) 2001/01/01

\*------------------------------------------------------------------------*/

/*************** Header files *********************************************/
#include <stdio.h>
#include <time.h>
#include "seed.h"

/*************** Assertions ***********************************************/

/*************** Definitions / Macros  ************************************/
#define KILO	1024
#define MEGA	(KILO*KILO)

////	Prine out BYTE data in ascending order and with no '0x'(hexa type)
#undef PrintBYTE
#define PrintBYTE(pfile, msg, Data, DataLen) {			\
	int idx;											\
	fprintf(pfile, "%5s =", msg);						\
	for( idx=0; idx<(int)DataLen; idx++) {				\
		if( (idx==0) || ((idx%16)!=0) )					\
			fprintf(pfile, " %.2x", Data[idx]);			\
		else											\
			fprintf(pfile, "\n\t%.2x", Data[idx]);		\
	}													\
	fprintf(pfile, "\n");								\
}

////	ѹ 'ByteLen'-bytes Ÿ óϴ  'Oper'
////	'Iter' ϰ,   KByte/sec .
#define SPEED_TEST(str1, str2, Iter, ByteLen, Oper) {	\
	unsigned idx;										\
	clock_t start, finish;								\
	double Sec, Mbps=0.0;								\
	start = clock();									\
	for( idx=0; idx<Iter; idx++)						\
		{	Oper;	}									\
	finish = clock();									\
	Sec = (double)(finish-start) / CLOCKS_PER_SEC;		\
	if( Sec!=0 )	Mbps = 8.0*ByteLen*idx/Sec/MEGA;	\
	printf("%s%7.3fMbps(=%d*", str1, Mbps, Iter);		\
	printf("%d", ByteLen);								\
	printf("/%3.0f)%s", 1000.0*Sec, str2);				\
}

/*************** New Data Types *******************************************/

/*************** Constant (Error Code) ************************************/
#define CTR_USAGE_ERROR		0x2001
#define CTR_KEYFILE_ERROR	0x2002

/*************** Global Variables *****************************************/
char	Help[] = "\
Usage1 : -[T/S]   (Test Value/Test Speed)\n\
Usage2 : -[E/D] -[ECB/CBC/OFB/CFB/ECBPAD/CBCPAD/OFBPAD/CFBPAD] infile outfile\n\
            (file 'key.dat' contains UserKey and IV)\n";

/*************** Prototypes ***********************************************/

//########################################################//
//######		Error Code  Լ				######//
//########################################################//

/*************************************************************************/
typedef struct{
	DWORD		ErrorCode;
	BYTE		Message[32];
} ERROR_MESSAGE;

ERROR_MESSAGE	ErrorMessage[] = {
	{CTR_FATAL_ERROR,		"CTR_FATAL_ERROR"},
	{CTR_INVALID_USERKEYLEN,"CTR_INVALID_USERKEYLEN"},
	{CTR_PAD_CHECK_ERROR,	"CTR_PAD_CHECK_ERROR"},
	{CTR_DATA_LEN_ERROR,	"CTR_DATA_LEN_ERROR"},
	{CTR_CIPHER_LEN_ERROR,	"CTR_CIPHER_LEN_ERROR"},
	{CTR_USAGE_ERROR,		"CTR_USAGE_ERROR"},
	{CTR_KEYFILE_ERROR,		"CTR_KEYFILE_ERROR"},
	{0, ""},
};

/**************************************************************************
*
*/
void	Error(
		DWORD	ErrorCode,
		char	*Message)
{
	DWORD	i;

	for( i=0; ErrorMessage[i].ErrorCode!=0; i++)
		if( ErrorMessage[i].ErrorCode==ErrorCode )	break;

	printf("ERROR(%s) :::: %s\n", ErrorMessage[i].Message, Message);
	if( ErrorCode==CTR_USAGE_ERROR )	printf("\n%s", Help);
	exit(1);
}

//########################################################//
//######		Validity Test  Լ				######//
//########################################################//

/*************************************************************************/
typedef struct{
	DWORD		ModeType;
	DWORD		PadType;
	char		Description[32];
	DWORD		UkLen;
	BYTE		UK[56];
	DWORD		IVLen;
	BYTE		IV[16];
	DWORD		PtLen;
	BYTE		PT[48];
	DWORD		EtLen;
	BYTE		ET[48];
} TEST_STRUCT;

TEST_STRUCT		TestData[] = {
	{	AI_ECB, AI_NO_PADDING, "ECB-NO_PADDING : ",
		16,		{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
				 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
		 0,		{0},
		16,		{0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,
				 0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F},
		16,		{0x5E,0xBA,0xC6,0xE0,0x05,0x4E,0x16,0x68,
				 0x19,0xAF,0xF1,0xCC,0x6D,0x34,0x6C,0xDB}	},
	{	AI_ECB, AI_NO_PADDING, "ECB-NO_PADDING",
		16,		{0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,
				 0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F},
		 0,		{0},
		16,		{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
				 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
		16,		{0xC1,0x1F,0x22,0xF2,0x01,0x40,0x50,0x50,
				 0x84,0x48,0x35,0x97,0xE4,0x37,0x0F,0x43}	},
	{	AI_ECB, AI_NO_PADDING, "ECB-NO_PADDING",
		16,		{0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF,
				 0x12,0x34,0x56,0x78,0x9A,0xBC,0xDE,0xF0},
		 0,		{0},
		32,		{0x37,0x36,0x35,0x34,0x33,0x32,0x31,0x20,
				 0x4E,0x6F,0x77,0x20,0x69,0x73,0x20,0x74,
				 0x68,0x65,0x20,0x74,0x69,0x6D,0x65,0x20,
				 0x66,0x6F,0x72,0x20,0x00,0x00,0x00,0x00},
		32,		{0x43,0x55,0x54,0x7A,0x75,0x1A,0xDC,0x6A,
				 0x2B,0xB6,0xC2,0xB1,0x88,0xF6,0x02,0xFB,
				 0xF5,0xC9,0x11,0x7A,0x23,0xCA,0xCC,0x7F,
				 0xA9,0xCA,0xBB,0xC5,0x32,0x9A,0x1C,0x86}	},
	{	AI_CBC, AI_NO_PADDING, "CBC-NO_PADDING",
		16,		{0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF,
				 0x12,0x34,0x56,0x78,0x9A,0xBC,0xDE,0xF0},
		 0,		{0},
		32,		{0x37,0x36,0x35,0x34,0x33,0x32,0x31,0x20,
				 0x4E,0x6F,0x77,0x20,0x69,0x73,0x20,0x74,
				 0x68,0x65,0x20,0x74,0x69,0x6D,0x65,0x20,
				 0x66,0x6F,0x72,0x20,0x00,0x00,0x00,0x00},
		32,		{0x43,0x55,0x54,0x7A,0x75,0x1A,0xDC,0x6A,
				 0x2B,0xB6,0xC2,0xB1,0x88,0xF6,0x02,0xFB,
				 0xFD,0x8F,0x32,0x28,0x85,0x4C,0xD5,0xFC,
				 0x33,0x6A,0x7C,0x2E,0xCA,0x89,0x02,0x18}	},
	{	AI_CBC, AI_PKCS_PADDING, "CBC-PKCS_PADDING",
		16,		{0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF,
				 0x12,0x34,0x56,0x78,0x9A,0xBC,0xDE,0xF0},
		16,		{0xFE,0xDC,0xBA,0x98,0x76,0x54,0x32,0x10,
				 0xFE,0xDC,0xBA,0x98,0x76,0x54,0x32,0x10},
		28,		"7654321 Now is the time for ",
		32,		{0xC7,0x67,0x4D,0x6A,0x8E,0x97,0xDC,0xD4,
				 0x12,0xBA,0xEA,0x21,0x0C,0x2B,0xAD,0x7F,
				 0x4A,0x3F,0x1B,0xFF,0xC2,0x57,0x2F,0x59,
				 0xB8,0x6B,0xE1,0x75,0x2C,0xAA,0xEB,0xEE}	},
	{	AI_OFB, 0, "OFB",
		16,		{0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF,
				 0x12,0x34,0x56,0x78,0x9A,0xBC,0xDE,0xF0},
		16,		{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
				 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
		28,		"7654321 Now is the time for ",
		28,		{0x07,0xDE,0x66,0xDE,0x7A,0xAE,0xF2,0xDE,
				 0xD0,0xC4,0x67,0x06,0x93,0x29,0x3E,0xE9,
				 0xE7,0xBA,0x9D,0x1C,0xAC,0xC9,0xDF,0x68,
				 0x4A,0x04,0x8C,0x1A}	},
	{	AI_CFB, 0, "CFB",
		16,		{0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF,
				 0x12,0x34,0x56,0x78,0x9A,0xBC,0xDE,0xF0},
		16,		{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
				 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
		28,		"7654321 Now is the time for ",
		32,		{0x07,0xDE,0x66,0xDE,0x7A,0xAE,0xF2,0xDE,
				 0xD0,0xC4,0x67,0x06,0x93,0x29,0x3E,0xE9,
				 0xE7,0xAA,0x43,0x7E,0x21,0x1D,0xC2,0x3E,
				 0x24,0x05,0x76,0xE1}	},
	{	0, 0, ""},
};

/**************************************************************************
*
*/
void	ValidityTest()
{
	BYTE	EncText[2*SEED_BLOCK_LEN], DecText[2*SEED_BLOCK_LEN];
	DWORD	i, tt, EncLen, DecLen;
	RET_VAL	ret;
	SEED_ALG_INFO	AlgInfo;

	//
	for( i=0; TestData[i].ModeType!=0; i++) {
		SEED_SetAlgInfo(TestData[i].ModeType, TestData[i].PadType,
						TestData[i].IV, &AlgInfo);

		//
		ret = SEED_KeySchedule(TestData[i].UK, TestData[i].UkLen, &AlgInfo);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_KeySchedule() returns.");

		//
		EncLen = tt = 0;
		ret = SEED_EncInit(&AlgInfo);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_EncInit() returns.");
		ret = SEED_EncUpdate(&AlgInfo, TestData[i].PT, TestData[i].PtLen,
							EncText, &EncLen);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_EncUpdate() returns.");
		ret = SEED_EncFinal(&AlgInfo, EncText+EncLen, &tt);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_EncFinal() returns.");
		EncLen += tt;

		//
		DecLen = tt = 0;
		ret = SEED_DecInit(&AlgInfo);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_DecInit() returns.");
		ret = SEED_DecUpdate(&AlgInfo, EncText, EncLen, DecText, &DecLen);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_DecUpdate() returns.");
		ret = SEED_DecFinal(&AlgInfo, DecText+DecLen, &tt);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_DecFinal() returns.");
		DecLen += tt;

		//
		fprintf(stdout, "\n==== test %d : %s\n", i, TestData[i].Description);
		PrintBYTE(stdout, "UK", TestData[i].UK, TestData[i].UkLen);
		PrintBYTE(stdout, "PT", TestData[i].PT, TestData[i].PtLen);
		PrintBYTE(stdout, "ET", EncText, EncLen);
		if( memcmp(TestData[i].PT, DecText, TestData[i].PtLen)==0 )
			printf("   DT = (OK)\n");
		else
			PrintBYTE(stdout, "DT", DecText, DecLen);
	}
}

/**************************************************************************
*
*/
void	SpeedTest()
{
#define DATA_LEN	1024
	BYTE	UserKey[SEED_USER_KEY_LEN], IV[SEED_BLOCK_LEN];
	BYTE	PT[DATA_LEN+32], ET[DATA_LEN+32], DT[DATA_LEN+32];
	DWORD	i, tt, Iter, EncLen=0, DecLen;
	RET_VAL	ret;
	SEED_ALG_INFO	AlgInfo;

	//
	for( tt=0; tt<SEED_USER_KEY_LEN; tt++)	UserKey[tt] = 0;
	for( tt=0; tt<SEED_BLOCK_LEN; tt++)		IV[tt] = 0;
	SEED_SetAlgInfo(AI_CBC, AI_NO_PADDING, IV, &AlgInfo);

	//
	ret = SEED_KeySchedule(UserKey, SEED_USER_KEY_LEN, &AlgInfo);
	if( ret!=CTR_SUCCESS )	Error(ret, "SEED_KeySchedule() returns.");

	for( i=0; i<3; i++) {
		fprintf(stdout, "\n==== test %d\n", i);
		Iter = 2 * 6500;

		//
		for( tt=0; tt<DATA_LEN; tt++)	PT[tt] = 0;

		//
		SPEED_TEST("Enc:", " / ", Iter, DATA_LEN,
			{	EncLen = tt = 0;
				ret = SEED_EncInit(&AlgInfo);
				ret = SEED_EncUpdate(&AlgInfo, PT, DATA_LEN,
									ET, &EncLen);
				ret = SEED_EncFinal(&AlgInfo, ET+EncLen, &tt);
				EncLen += tt;									}	);
		//
		SPEED_TEST("Dec:", " // ", Iter, DATA_LEN,
			{	DecLen = tt = 0;
				ret = SEED_DecInit(&AlgInfo);
				ret = SEED_DecUpdate(&AlgInfo, ET, EncLen,
									DT, &DecLen);
				ret = SEED_DecFinal(&AlgInfo, DT+DecLen, &tt);
				DecLen += tt;									}	);

		//
		for( tt=0; tt<DATA_LEN; tt++)
			if( DT[tt]!=0 )	break;
		if( tt==DATA_LEN )	printf("OK");
		else				printf("FAIL");
	}
#undef DATA_LEN
}

/**************************************************************************
*
*/
void	GetKeyIV(
		BYTE	Key[SEED_USER_KEY_LEN],
		DWORD	*KeyLen,
		BYTE	IV[SEED_BLOCK_LEN],
		DWORD	*IVLen)
{
	DWORD	ch, i, j;
	FILE	*pfile;

	if( (pfile=fopen("key.dat", "r"))==NULL ) {
		printf( "The file 'key.dat' was not opened\n");
		Error(0, "File(key.dat) Open Error");
	}

	fscanf(pfile, "%d", &j);
	*KeyLen = j;
	for( i=0; i<j; i++) {
		fscanf(pfile, "%X", &ch);
		Key[i] = (BYTE) ch;
	}

	fscanf(pfile, "%d", &j);
	if( j!=SEED_BLOCK_LEN ) {
		*IVLen = 0;
		for( i=0; i<SEED_BLOCK_LEN; i++)
			IV[i] = (BYTE) 0;
	}
	else {
		*IVLen = SEED_BLOCK_LEN;
		for( i=0; i<SEED_BLOCK_LEN; i++) {
			fscanf(pfile, "%X", &ch);
			IV[i] = (BYTE) ch;
		}
	}

	fclose(pfile);
}

/**************************************************************************
*
*/
void	GeneralTest(
		DWORD	EncType,
		DWORD	ModeType,
		DWORD	PadType,
		char	*infile,
		char	*outfile)
{
	FILE	*pIn, *pOut;
	BYTE	UserKey[SEED_USER_KEY_LEN];
	BYTE	IV[SEED_BLOCK_LEN];
	BYTE	SrcData[1024+32], DstData[1024+32];
	DWORD	UKLen, IVLen, SrcLen, DstLen;
	RET_VAL	ret;
	SEED_ALG_INFO	AlgInfo;

	//
	GetKeyIV(UserKey, &UKLen, IV, &IVLen);

	//
	if( (pIn=fopen(infile, "rb"))==NULL ) {
		printf( "The file '%s' was not opened\n", infile);
		Error(0, "File(infile) Open Error");
	}
	if( (pOut=fopen(outfile, "wb"))==NULL ) {
		printf( "The file '%s' was not opened\n", outfile);
		Error(0, "File(outfile) Open Error");
	}
	//
	SEED_SetAlgInfo(ModeType, PadType, IV, &AlgInfo);
	ret = SEED_KeySchedule(UserKey, UKLen, &AlgInfo);
	if( ret!=CTR_SUCCESS )	Error(ret, "SEED_KeySchedule() returns.");

	if( EncType==0 ) {		//	Encryption
		ret = SEED_EncInit(&AlgInfo);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_EncInit() returns.");

		for(  ;  ;  ) {
			SrcLen = fread(SrcData, sizeof(BYTE), 1024, pIn);
			if( SrcLen==0 )	break;

			DstLen = 1024;
			ret = SEED_EncUpdate(&AlgInfo, SrcData, SrcLen, DstData, &DstLen);
			if( ret!=CTR_SUCCESS )	Error(ret, "SEED_EncUpdate() returns.");

			fwrite(DstData, sizeof(BYTE), DstLen, pOut);
		}

		DstLen = 1024;
		ret = SEED_EncFinal(&AlgInfo, DstData, &DstLen);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_EncFinal() returns.");
		fwrite(DstData, sizeof(BYTE), DstLen, pOut);
	}
	else {					//	Decryption
		ret = SEED_DecInit(&AlgInfo);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_DecInit() returns.");

		for(  ;  ;  ) {
			SrcLen = fread(SrcData, sizeof(BYTE), 1024, pIn);
			if( SrcLen==0 )	break;

			DstLen = 1024;
			ret = SEED_DecUpdate(&AlgInfo, SrcData, SrcLen, DstData, &DstLen);
			if( ret!=CTR_SUCCESS )	Error(ret, "SEED_DecUpdate() returns.");

			fwrite(DstData, sizeof(BYTE), DstLen, pOut);
		}

		DstLen = 1024;
		ret = SEED_DecFinal(&AlgInfo, DstData, &DstLen);
		if( ret!=CTR_SUCCESS )	Error(ret, "SEED_DecFinal() returns.");
		fwrite(DstData, sizeof(BYTE), DstLen, pOut);
	}

	fclose(pIn);
	fclose(pOut);
}

/**************************************************************************
*
*/
void	ChangeToCapital(char *String)
{
	DWORD	i;

	if( String[0]=='-' ) {
		for( i=1;  ; i++) {
			if( String[i]=='\0' )	break;
			if( (String[i]>='a') && (String[i]<='z') )
				String[i] += 'A' - 'a';
		}
	}
}

/**************************************************************************
*
*/
void	main(int argc, char **argv)
{
	DWORD	i;

	//
	for( i=1; i<(DWORD)argc; i++)
		ChangeToCapital(argv[i]);

	//
	if( argc==2 ) {
		if( strcmp(argv[1], "-T")==0 )		ValidityTest();
		else if( strcmp(argv[1], "-S")==0 )	SpeedTest();
		else	Error(CTR_USAGE_ERROR, "Invalid Use of Argument");
	}
	else if( argc==5 ) {
		DWORD	EncType=0, ModeType=0, PadType=0;

		//
		if( strcmp(argv[1], "-E")==0 )		EncType = 0;
		else if( strcmp(argv[1], "-D")==0 )	EncType = 1;
		else	Error(CTR_USAGE_ERROR, "Invalid Use of Argument");

		//
		if( strcmp(argv[2], "-ECB")==0 ) {
			ModeType = AI_ECB;
			PadType = AI_NO_PADDING;
		}
		else if( strcmp(argv[2], "-CBC")==0 ) {
			ModeType = AI_CBC;
			PadType = AI_NO_PADDING;
		}
		else if( strcmp(argv[2], "-OFB")==0 ) {
			ModeType = AI_OFB;
			PadType = AI_NO_PADDING;
		}
		else if( strcmp(argv[2], "-CFB")==0 ) {
			ModeType = AI_CFB;
			PadType = AI_NO_PADDING;
		}
		else if( strcmp(argv[2], "-ECBPAD")==0 ) {
			ModeType = AI_ECB;
			PadType = AI_PKCS_PADDING;
		}
		else if( strcmp(argv[2], "-CBCPAD")==0 ) {
			ModeType = AI_CBC;
			PadType = AI_PKCS_PADDING;
		}
		else if( strcmp(argv[2], "-OFBPAD")==0 ) {
			ModeType = AI_OFB;
			PadType = AI_PKCS_PADDING;
		}
		else if( strcmp(argv[2], "-CFBPAD")==0 ) {
			ModeType = AI_CFB;
			PadType = AI_PKCS_PADDING;
		}
		else	Error(CTR_USAGE_ERROR, "Invalid Use of Argument");

		GeneralTest(EncType, ModeType, PadType, argv[3], argv[4]);
	}
	else
		Error(CTR_USAGE_ERROR, "Invalid Number of Argument");
}

/*************** END OF FILE **********************************************/
